@@ -129,12 +129,12 @@ test_ssm_completed_callback (FpiSsm   *ssm,
 }
 
 static FpiSsm *
-ssm_test_new_full (int nr_states, const char *name)
+ssm_test_new_full (int nr_states, int cleanup_state, const char *name)
 {
   FpiSsm *ssm;
   FpiSsmTestData *data;
 
-  ssm = fpi_ssm_new_full (fake_device, test_ssm_handler, nr_states, name);
+  ssm = fpi_ssm_new_full (fake_device, test_ssm_handler, nr_states, cleanup_state, name);
   data = fpi_ssm_test_data_new ();
   data->expected_last_state = nr_states;
   fpi_ssm_set_data (ssm, data, (GDestroyNotify) fpi_ssm_test_data_unref_by_ssm);
@@ -145,7 +145,7 @@ ssm_test_new_full (int nr_states, const char *name)
 static FpiSsm *
 ssm_test_new (void)
 {
-  return ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SSM");
+  return ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SSM");
 }
 
 static gboolean
@@ -158,14 +158,6 @@ test_ssm_cancel_delayed_action_delayed (gpointer data)
   return G_SOURCE_REMOVE;
 }
 
-static gboolean
-test_ssm_cancel_cancellable_delayed (gpointer data)
-{
-  g_cancellable_cancel (G_CANCELLABLE (data));
-
-  return G_SOURCE_REMOVE;
-}
-
 /* Tests */
 
 static void
@@ -188,7 +180,8 @@ test_ssm_new_full (void)
   FpiSsm *ssm;
 
   ssm = fpi_ssm_new_full (fake_device, test_ssm_handler,
-                          FPI_TEST_SSM_STATE_NUM, "Test SSM Name");
+                          FPI_TEST_SSM_STATE_NUM, FPI_TEST_SSM_STATE_NUM,
+                          "Test SSM Name");
 
   g_assert_null (fpi_ssm_get_data (ssm));
   g_assert_no_error (fpi_ssm_get_error (ssm));
@@ -215,6 +208,8 @@ test_ssm_new_wrong_states (void)
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
                          "*BUG:*nr_states*");
+  g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL,
+                         "*BUG:*start_cleanup*");
   ssm = fpi_ssm_new (fake_device, test_ssm_handler, -1);
   g_test_assert_expected_messages ();
 }
@@ -271,7 +266,7 @@ test_ssm_start_single (void)
   g_autoptr(FpiSsmTestData) data = NULL;
   FpiSsm *ssm;
 
-  ssm = ssm_test_new_full (1, "FPI_TEST_SSM_SINGLE_STATE");
+  ssm = ssm_test_new_full (1, 1, "FPI_TEST_SSM_SINGLE_STATE");
   data = fpi_ssm_test_data_ref (fpi_ssm_get_data (ssm));
 
   fpi_ssm_start (ssm, test_ssm_completed_callback);
@@ -336,7 +331,7 @@ test_ssm_next_with_delayed (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
   fpi_ssm_next_state (ssm);
@@ -444,7 +439,7 @@ test_ssm_jump_to_state_with_delayed (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
   fpi_ssm_jump_to_state (ssm, FPI_TEST_SSM_STATE_2);
@@ -561,7 +556,7 @@ test_ssm_mark_completed_with_delayed (void)
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
   data->expected_last_state = FPI_TEST_SSM_STATE_0;
-  fpi_ssm_mark_completed_delayed (ssm, 10, NULL);
+  fpi_ssm_mark_completed_delayed (ssm, 10);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
   fpi_ssm_mark_completed (g_steal_pointer (&ssm));
@@ -623,7 +618,7 @@ test_ssm_mark_failed_with_delayed (void)
   fpi_ssm_start (ssm, test_ssm_completed_callback);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_mark_completed_delayed (ssm, 10, NULL);
+  fpi_ssm_mark_completed_delayed (ssm, 10);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
   data->expected_last_state = FPI_TEST_SSM_STATE_0;
@@ -654,7 +649,7 @@ test_ssm_delayed_next (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
@@ -682,7 +677,7 @@ test_ssm_delayed_next_cancel (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
@@ -701,40 +696,6 @@ test_ssm_delayed_next_cancel (void)
   g_assert_no_error (data->error);
 }
 
-static void
-test_ssm_delayed_next_cancellable (void)
-{
-  g_autoptr(FpiSsm) ssm = ssm_test_new ();
-  g_autoptr(GCancellable) cancellable = g_cancellable_new ();
-  FpiSsmTestData *data = fpi_ssm_get_data (ssm);
-
-  fpi_ssm_start (ssm, test_ssm_completed_callback);
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  fpi_ssm_next_state_delayed (ssm, 10, cancellable);
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  g_idle_add_full (G_PRIORITY_HIGH, test_ssm_cancel_cancellable_delayed, cancellable, NULL);
-
-  while (!g_cancellable_is_cancelled (cancellable))
-    g_main_context_iteration (NULL, TRUE);
-
-  g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
-  fpi_ssm_cancel_delayed_state_change (ssm);
-  g_test_assert_expected_messages ();
-
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  g_assert_false (data->completed);
-  g_assert_no_error (data->error);
-}
-
 static void
 test_ssm_delayed_next_not_started (void)
 {
@@ -742,7 +703,7 @@ test_ssm_delayed_next_not_started (void)
   FpiSsmTestData *data = fpi_ssm_get_data (ssm);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*completed*");
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
   g_test_assert_expected_messages ();
 
   g_assert_cmpint (data->handler_state, ==, -1);
@@ -773,7 +734,7 @@ test_ssm_delayed_next_complete (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
@@ -785,7 +746,7 @@ test_ssm_delayed_next_complete (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_1);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 2);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_1);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_1);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 2);
@@ -797,7 +758,7 @@ test_ssm_delayed_next_complete (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_2);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 3);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_2);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_2);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 3);
@@ -809,7 +770,7 @@ test_ssm_delayed_next_complete (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_3);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 4);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_3);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_3);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 4);
@@ -835,7 +796,7 @@ test_ssm_delayed_jump_to_state (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10);
 
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
@@ -848,7 +809,7 @@ test_ssm_delayed_jump_to_state (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_2);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 2);
 
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_1, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_1, 10);
 
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_2);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_2);
@@ -877,7 +838,7 @@ test_ssm_delayed_jump_to_state_cancel (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
@@ -896,40 +857,6 @@ test_ssm_delayed_jump_to_state_cancel (void)
   g_assert_no_error (data->error);
 }
 
-static void
-test_ssm_delayed_jump_to_state_cancellable (void)
-{
-  g_autoptr(FpiSsm) ssm = ssm_test_new ();
-  g_autoptr(GCancellable) cancellable = g_cancellable_new ();
-  FpiSsmTestData *data = fpi_ssm_get_data (ssm);
-
-  fpi_ssm_start (ssm, test_ssm_completed_callback);
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10, cancellable);
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  g_idle_add_full (G_PRIORITY_HIGH, test_ssm_cancel_cancellable_delayed, cancellable, NULL);
-
-  while (!g_cancellable_is_cancelled (cancellable))
-    g_main_context_iteration (NULL, TRUE);
-
-  g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
-  fpi_ssm_cancel_delayed_state_change (ssm);
-  g_test_assert_expected_messages ();
-
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  g_assert_false (data->completed);
-  g_assert_no_error (data->error);
-}
-
 static void
 test_ssm_delayed_jump_to_state_not_started (void)
 {
@@ -937,7 +864,7 @@ test_ssm_delayed_jump_to_state_not_started (void)
   FpiSsmTestData *data = fpi_ssm_get_data (ssm);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*completed*");
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_2, 10);
   g_test_assert_expected_messages ();
 
   g_assert_cmpint (data->handler_state, ==, -1);
@@ -967,7 +894,7 @@ test_ssm_delayed_jump_to_state_last (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_3, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_3, 10);
 
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
@@ -996,7 +923,7 @@ test_ssm_delayed_jump_to_state_wrong (void)
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*nr_states*");
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_NUM + 10, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_NUM + 10, 10);
   g_test_assert_expected_messages ();
 
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
@@ -1013,7 +940,7 @@ test_ssm_delayed_jump_to_state_wrong (void)
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 2);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*state*");
-  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_0 - 10, 10, NULL);
+  fpi_ssm_jump_to_state_delayed (ssm, FPI_TEST_SSM_STATE_0 - 10, 10);
   g_test_assert_expected_messages ();
 
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_NUM + 10);
@@ -1043,7 +970,7 @@ test_ssm_delayed_mark_completed (void)
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
   data->expected_last_state = FPI_TEST_SSM_STATE_0;
-  fpi_ssm_mark_completed_delayed (g_steal_pointer (&ssm), 10, NULL);
+  fpi_ssm_mark_completed_delayed (g_steal_pointer (&ssm), 10);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
   while (g_slist_length (data->handlers_chain) == 1)
@@ -1063,7 +990,7 @@ test_ssm_delayed_mark_completed_not_started (void)
   g_autoptr(FpiSsmTestData) data = fpi_ssm_test_data_ref (fpi_ssm_get_data (ssm));
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*completed*");
-  fpi_ssm_mark_completed_delayed (ssm, 10, NULL);
+  fpi_ssm_mark_completed_delayed (ssm, 10);
   g_test_assert_expected_messages ();
 
   g_timeout_add (100, G_SOURCE_FUNC (fpi_ssm_test_nullify_pointer), &ssm);
@@ -1090,7 +1017,7 @@ test_ssm_delayed_mark_completed_cancel (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_mark_completed_delayed (ssm, 10, NULL);
+  fpi_ssm_mark_completed_delayed (ssm, 10);
   g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
@@ -1110,40 +1037,6 @@ test_ssm_delayed_mark_completed_cancel (void)
   g_assert_false (data->ssm_destroyed);
 }
 
-static void
-test_ssm_delayed_mark_completed_cancellable (void)
-{
-  g_autoptr(FpiSsm) ssm = ssm_test_new ();
-  g_autoptr(GCancellable) cancellable = g_cancellable_new ();
-  FpiSsmTestData *data = fpi_ssm_get_data (ssm);
-
-  fpi_ssm_start (ssm, test_ssm_completed_callback);
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  fpi_ssm_mark_completed_delayed (ssm, 10, cancellable);
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  g_idle_add_full (G_PRIORITY_HIGH, test_ssm_cancel_cancellable_delayed, cancellable, NULL);
-
-  while (!g_cancellable_is_cancelled (cancellable))
-    g_main_context_iteration (NULL, TRUE);
-
-  g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
-  fpi_ssm_cancel_delayed_state_change (ssm);
-  g_test_assert_expected_messages ();
-
-  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
-  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
-
-  g_assert_false (data->completed);
-  g_assert_no_error (data->error);
-}
-
 static void
 test_ssm_delayed_cancel_error (void)
 {
@@ -1170,7 +1063,7 @@ test_ssm_subssm_start (void)
 {
   g_autoptr(FpiSsm) ssm = ssm_test_new ();
   g_autoptr(FpiSsm) subssm =
-    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
+    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
   FpiSsmTestData *data = fpi_ssm_get_data (ssm);
 
   g_autoptr(FpiSsmTestData) subdata =
@@ -1222,7 +1115,7 @@ test_ssm_subssm_mark_failed (void)
 {
   g_autoptr(FpiSsm) ssm = ssm_test_new ();
   g_autoptr(FpiSsm) subssm =
-    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
+    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
   g_autoptr(FpiSsmTestData) data =
     fpi_ssm_test_data_ref (fpi_ssm_get_data (ssm));
   g_autoptr(FpiSsmTestData) subdata =
@@ -1261,7 +1154,7 @@ test_ssm_subssm_start_with_started (void)
 {
   g_autoptr(FpiSsm) ssm = ssm_test_new ();
   g_autoptr(FpiSsm) subssm =
-    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
+    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
   FpiSsmTestData *data = fpi_ssm_get_data (ssm);
 
   g_autoptr(FpiSsmTestData) subdata =
@@ -1305,7 +1198,7 @@ test_ssm_subssm_start_with_delayed (void)
 {
   g_autoptr(FpiSsm) ssm = ssm_test_new ();
   g_autoptr(FpiSsm) subssm =
-    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
+    ssm_test_new_full (FPI_TEST_SSM_STATE_NUM, FPI_TEST_SSM_STATE_NUM, "FPI_TEST_SUB_SSM");
   FpiSsmTestData *data = fpi_ssm_get_data (ssm);
 
   g_autoptr(FpiSsmTestData) subdata =
@@ -1317,7 +1210,7 @@ test_ssm_subssm_start_with_delayed (void)
   g_assert_cmpint (fpi_ssm_get_cur_state (ssm), ==, FPI_TEST_SSM_STATE_0);
   g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
 
-  fpi_ssm_next_state_delayed (ssm, 10, NULL);
+  fpi_ssm_next_state_delayed (ssm, 10);
 
   g_test_expect_message (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, "*BUG:*timeout*");
   fpi_ssm_start_subsm (ssm, subssm);
@@ -1348,6 +1241,76 @@ test_ssm_subssm_start_with_delayed (void)
   g_assert_no_error (data->error);
 }
 
+static void
+test_ssm_cleanup_complete (void)
+{
+  FpiSsm *ssm = ssm_test_new_full (4, FPI_TEST_SSM_STATE_2, "FPI_TEST_SSM");
+
+  g_autoptr(FpiSsmTestData) data = fpi_ssm_test_data_ref (fpi_ssm_get_data (ssm));
+
+  fpi_ssm_start (ssm, test_ssm_completed_callback);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
+
+  data->expected_last_state = FPI_TEST_SSM_STATE_3;
+
+  /* Completing jumps to the cleanup state */
+  fpi_ssm_mark_completed (ssm);
+  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_2);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 2);
+  g_assert_false (data->completed);
+  g_assert_false (data->ssm_destroyed);
+
+  /* Completing again jumps to the next cleanup state */
+  fpi_ssm_mark_completed (ssm);
+  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_3);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 3);
+  g_assert_false (data->completed);
+  g_assert_false (data->ssm_destroyed);
+
+  /* Completing again finalizes everything */
+  fpi_ssm_mark_completed (ssm);
+  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_3);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 4);
+  g_assert_true (data->completed);
+  g_assert_no_error (data->error);
+  g_assert_true (data->ssm_destroyed);
+}
+
+static void
+test_ssm_cleanup_fail (void)
+{
+  FpiSsm *ssm = ssm_test_new_full (4, FPI_TEST_SSM_STATE_2, "FPI_TEST_SSM");
+
+  g_autoptr(FpiSsmTestData) data = fpi_ssm_test_data_ref (fpi_ssm_get_data (ssm));
+
+  fpi_ssm_start (ssm, test_ssm_completed_callback);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 1);
+
+  data->expected_last_state = FPI_TEST_SSM_STATE_3;
+
+  /* Failing jumps to the cleanup state */
+  fpi_ssm_mark_failed (ssm, g_error_new (G_IO_ERROR, G_IO_ERROR_CANCELLED, "non-cleanup"));
+  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_2);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 2);
+  g_assert_false (data->completed);
+  g_assert_false (data->ssm_destroyed);
+
+  /* Failing again jumps to the next cleanup state */
+  fpi_ssm_mark_failed (ssm, g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED, "cleanup 1"));
+  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_3);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 3);
+  g_assert_false (data->completed);
+  g_assert_false (data->ssm_destroyed);
+
+  /* Failing again finalizes everything */
+  fpi_ssm_mark_failed (ssm, g_error_new (G_IO_ERROR, G_IO_ERROR_FAILED, "cleanup 2"));
+  g_assert_cmpint (data->handler_state, ==, FPI_TEST_SSM_STATE_3);
+  g_assert_cmpuint (g_slist_length (data->handlers_chain), ==, 4);
+  g_assert_true (data->completed);
+  g_assert_error (data->error, G_IO_ERROR, G_IO_ERROR_CANCELLED);
+  g_assert_true (data->ssm_destroyed);
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -1384,24 +1347,23 @@ main (int argc, char *argv[])
   g_test_add_func ("/ssm/mark_failed/with_delayed", test_ssm_mark_failed_with_delayed);
   g_test_add_func ("/ssm/delayed/next", test_ssm_delayed_next);
   g_test_add_func ("/ssm/delayed/next/cancel", test_ssm_delayed_next_cancel);
-  g_test_add_func ("/ssm/delayed/next/cancellable", test_ssm_delayed_next_cancellable);
   g_test_add_func ("/ssm/delayed/next/not_started", test_ssm_delayed_next_not_started);
   g_test_add_func ("/ssm/delayed/next/complete", test_ssm_delayed_next_complete);
   g_test_add_func ("/ssm/delayed/jump_to_state", test_ssm_delayed_jump_to_state);
   g_test_add_func ("/ssm/delayed/jump_to_state/cancel", test_ssm_delayed_jump_to_state_cancel);
-  g_test_add_func ("/ssm/delayed/jump_to_state/cancellable", test_ssm_delayed_jump_to_state_cancellable);
   g_test_add_func ("/ssm/delayed/jump_to_state/not_started", test_ssm_delayed_jump_to_state_not_started);
   g_test_add_func ("/ssm/delayed/jump_to_state/last", test_ssm_delayed_jump_to_state_last);
   g_test_add_func ("/ssm/delayed/jump_to_state/wrong", test_ssm_delayed_jump_to_state_wrong);
   g_test_add_func ("/ssm/delayed/mark_completed", test_ssm_delayed_mark_completed);
   g_test_add_func ("/ssm/delayed/mark_completed/cancel", test_ssm_delayed_mark_completed_cancel);
-  g_test_add_func ("/ssm/delayed/mark_completed/cancellable", test_ssm_delayed_mark_completed_cancellable);
   g_test_add_func ("/ssm/delayed/mark_completed/not_started", test_ssm_delayed_mark_completed_not_started);
   g_test_add_func ("/ssm/delayed/cancel/error", test_ssm_delayed_cancel_error);
   g_test_add_func ("/ssm/subssm/start", test_ssm_subssm_start);
   g_test_add_func ("/ssm/subssm/start/with_started", test_ssm_subssm_start_with_started);
   g_test_add_func ("/ssm/subssm/start/with_delayed", test_ssm_subssm_start_with_delayed);
   g_test_add_func ("/ssm/subssm/mark_failed", test_ssm_subssm_mark_failed);
+  g_test_add_func ("/ssm/cleanup/complete", test_ssm_cleanup_complete);
+  g_test_add_func ("/ssm/cleanup/fail", test_ssm_cleanup_fail);
 
   return g_test_run ();
 }
